/*!\file window.c
 *
 * \brief Utilisation de la SDL2 et d'OpenGL 3+ - flou, sobel + itrations de Flous en GPU
 * \author Fars BELHADJ, amsi@ai.univ-paris8.fr
 * \date April 24 2014
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#include <gl4du.h>

/*
 * Prototypes des fonctions statiques contenues dans ce fichier C
 */
static SDL_Window * initWindow(int w, int h, SDL_GLContext * poglContext);
static void quit(void);
static void initGL(SDL_Window * win);
static void initData(void);
static void resizeGL(SDL_Window * win);
static void loop(SDL_Window * win);
static void manageEvents(SDL_Window * win);
static void draw(void);
static void printFPS(void);
static void myTex(GLuint * id);

/*!\brief dimensions de la fentre */
static GLfloat _dim[] = {800, 600};
/*!\brief pointeur vers la (future) fentre SDL */
static SDL_Window * _win = NULL;
/*!\brief pointeur vers le (futur) contexte OpenGL */
static SDL_GLContext _oglContext = NULL;
/*!\brief identifiant du (futur) vertex array object */
static GLuint _vao = 0;
/*!\brief identifiant du (futur) buffer de data */
static GLuint _buffer = 0;
/*!\brief identifiants des (futurs) GLSL programs */
static GLuint _pId[3] = {0, 0, 0};
/*!\brief indice de l'effet en cours : 0 rien, 1 blur, 2 sobel, 3 blur
 * itratif 10 fois, 4 blur itratif */
static GLuint _currEffect = 0;
/*!\brief identifiant de la texture */
static GLuint _tId = 0;
/*!\brief l'image charge  partir d'SDL */
static SDL_Surface * _image = NULL;

/*!\brief La fonction principale initialise la bibliothque SDL2,
 * demande la cration de la fentre SDL et du contexte OpenGL par
 * l'appel  \ref initWindow, initialise OpenGL avec \ref initGL et
 * lance la boucle (infinie) principale.
 */
int main(int argc, char ** argv) {
  if(SDL_Init(SDL_INIT_VIDEO) < 0) {
    fprintf(stderr, "Erreur lors de l'initialisation de SDL :  %s", SDL_GetError());
    return -1;
  }
  atexit(SDL_Quit);
  if((_win = initWindow(_dim[0], _dim[1], &_oglContext))) {
    initGL(_win);
    _pId[0] = gl4duCreateProgram("<vs>shaders/basic.vs", "<fs>shaders/basic.fs", NULL);
    _pId[1] = gl4duCreateProgram("<vs>shaders/basic.vs", "<fs>shaders/blur.fs", NULL);
    _pId[2] = gl4duCreateProgram("<vs>shaders/basic.vs", "<fs>shaders/sobel.fs", NULL);
    initData();
    loop(_win);
  }
  return 0;
}

/*!\brief Cette fonction est appele au moment de sortir du programme
 *  (atexit), elle libre la fentre SDL \ref _win et le contexte
 *  OpenGL \ref _oglContext.
 */
static void quit(void) {
  if(_vao)
    glDeleteVertexArrays(1, &_vao);
  if(_buffer)
    glDeleteBuffers(1, &_buffer);
  if(_buffer)
    glDeleteTextures(1, &_tId);
  if(_image)
    SDL_FreeSurface(_image);
  if(_oglContext)
    SDL_GL_DeleteContext(_oglContext);
  if(_win)
    SDL_DestroyWindow(_win);
  gl4duClean(GL4DU_ALL);
}

/*!\brief Cette fonction cr la fentre SDL de largeur \a w et de
 *  hauteur \a h, le contexte OpenGL \ref et stocke le pointeur dans
 *  poglContext. Elle retourne le pointeur vers la fentre SDL.
 *
 * Le contexte OpenGL cr est en version 3 pour
 * SDL_GL_CONTEXT_MAJOR_VERSION, en version 2 pour
 * SDL_GL_CONTEXT_MINOR_VERSION et en SDL_GL_CONTEXT_PROFILE_CORE
 * concernant le profile. Le double buffer est activ et le buffer de
 * profondeur est en 24 bits.
 *
 * \param w la largeur de la fentre  crer.
 * \param h la hauteur de la fentre  crer.
 * \param poglContext le pointeur vers la case o sera rfrenc le
 * contexte OpenGL cr.
 * \return le pointeur vers la fentre SDL si tout se passe comme
 * prvu, NULL sinon.
 */
static SDL_Window * initWindow(int w, int h, SDL_GLContext * poglContext) {
  SDL_Window * win = NULL;
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
  if( (win = SDL_CreateWindow("Fenetre GL", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 
			      w, h, SDL_WINDOW_OPENGL | SDL_WINDOW_RESIZABLE | 
			      SDL_WINDOW_SHOWN)) == NULL )
    return NULL;
  if( (*poglContext = SDL_GL_CreateContext(win)) == NULL ) {
    SDL_DestroyWindow(win);
    return NULL;
  }
  fprintf(stderr, "Version d'OpenGL : %s\n", glGetString(GL_VERSION));
  fprintf(stderr, "Version de shaders supportes : %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));  
  atexit(quit);
  return win;
}

/*!\brief Cette fonction initialise les paramtres OpenGL.
 *
 * \param win le pointeur vers la fentre SDL pour laquelle nous avons
 * attach le contexte OpenGL.
 */
static void initGL(SDL_Window * win) {
  glClearColor(0.2f, 0.2f, 0.2f, 0.0f);
  resizeGL(win);
}

static void initData(void) {
  SDL_Surface * t;
  GLfloat data[] = {
    /* 4 coordonnes de sommets */
    -1.f, -1.f, 1.f, -1.f,
    -1.f,  1.f, 1.f,  1.f,
    /* 2 coordonnes de texture par sommet */
    0.0f, 0.0f, 1.0f, 0.0f, 
    0.0f, 1.0f, 1.0f, 1.0f
  };
  glGenVertexArrays(1, &_vao);
  glBindVertexArray(_vao);

  glGenBuffers(1, &_buffer);
  glBindBuffer(GL_ARRAY_BUFFER, _buffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof data, data, GL_STATIC_DRAW);
  glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, (const void *)0);  
  glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, (const void *)((4 * 2) * sizeof *data));
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  glBindVertexArray(0);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  if( (t = SDL_LoadBMP("ciel.bmp")) != NULL ) {
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, t->w, t->h, 0, GL_BGRA, GL_UNSIGNED_BYTE, t->pixels);
    SDL_FreeSurface(t);
  } else {
    fprintf(stderr, "Erreur lors du chargement de la texture\n");
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 1, 1, 0, GL_BGRA, GL_UNSIGNED_BYTE, NULL);
  }
  glBindTexture(GL_TEXTURE_2D, 0);
}

/*!\brief Cette fonction paramtrela vue (viewPort) OpenGL en fonction
 * des dimensions de la fentre SDL pointe par \a win.
 *
 * \param win le pointeur vers la fentre SDL pour laquelle nous avons
 * attach le contexte OpenGL.
 */
static void resizeGL(SDL_Window * win) {
  int w, h;
  SDL_GetWindowSize(win, &w, &h);
  glViewport(0, 0, w, h);
}

/*!\brief Boucle infinie principale : gre les vnements, dessine,
 * imprime le FPS et swap les buffers.
 *
 * \param win le pointeur vers la fentre SDL pour laquelle nous avons
 * attach le contexte OpenGL.
 */
static void loop(SDL_Window * win) {
  SDL_GL_SetSwapInterval(1);
  for(;;) {
    manageEvents(win);
    draw();
    printFPS();
    SDL_GL_SwapWindow(win);
    gl4duUpdateShaders();
  }
}

/*!\brief Cette fonction permet de grer les vnements clavier et
 * souris via la bibliothque SDL et pour la fentre pointe par \a
 * win.
 *
 * \param win le pointeur vers la fentre SDL pour laquelle nous avons
 * attach le contexte OpenGL.
 */
static void manageEvents(SDL_Window * win) {
  SDL_Event event;
  while(SDL_PollEvent(&event)) 
    switch (event.type) {
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case 'n':
	_currEffect = (_currEffect + 1) % 5;
	break;
      case 'p':
	_currEffect = (_currEffect + 4) % 5;
	break;
      case SDLK_ESCAPE:
      case 'q':
	exit(0);
      default:
	fprintf(stderr, "La touche %s a ete pressee\n",
		SDL_GetKeyName(event.key.keysym.sym));
	break;
      }
      break;
    case SDL_KEYUP:
      break;
    case SDL_WINDOWEVENT:
      if(event.window.windowID == SDL_GetWindowID(win)) {
	switch (event.window.event)  {
	case SDL_WINDOWEVENT_RESIZED:
	  resizeGL(win);
	  break;
	case SDL_WINDOWEVENT_CLOSE:
	  event.type = SDL_QUIT;
	  SDL_PushEvent(&event);
	  break;
	}
      }
      break;
    case SDL_QUIT:
      exit(0);
    }
}

static void draw(void) {
  int i;
  static int ft = 1, nbIt4Effect4 = 0;
  static GLuint fbo, tex[2], pId;
  if(ft) {
    myTex(&tex[0]); glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, _dim[0], _dim[1], 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    myTex(&tex[1]); glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, _dim[0], _dim[1], 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glGenFramebuffers(1, &fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    ft = 0;
  }

  switch(_currEffect) {
  case 0: case 1: case 2: pId = _pId[_currEffect]; break;
  case 3: case 4: pId = _pId[1]; break;
  default: break;
  }

  glBindFramebuffer(GL_FRAMEBUFFER, fbo);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D,  tex[0],  0);
  glDisable(GL_DEPTH);

  glBindVertexArray(_vao);
  glEnableVertexAttribArray(0);
  glEnableVertexAttribArray(1); 
  glActiveTexture(GL_TEXTURE0);
  glUseProgram(pId);

  //if(_currEffect < 4)
  //  nbIt4Effect4 = 0;
  if(_currEffect < 4 || !nbIt4Effect4) {
    if(!nbIt4Effect4)
      nbIt4Effect4 = 1;
    else
      nbIt4Effect4 = 0;
    glBindTexture(GL_TEXTURE_2D, _tId);
    glUniform1i(glGetUniformLocation(pId, "myTexture"), 0);
    glUniform1f(glGetUniformLocation(pId, "width"), _dim[0]);
    glUniform1f(glGetUniformLocation(pId, "height"), _dim[1]);
    glUniform1i(glGetUniformLocation(pId, "inv"), 1);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
  }
  if(_currEffect == 3) {
    for(i = 0; i < 10; i++) {
      glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D,  tex[(i + 1) % 2],  0);
      glBindTexture(GL_TEXTURE_2D, tex[i % 2]);
      glUniform1i(glGetUniformLocation(pId, "myTexture"), 0);
      glUniform1f(glGetUniformLocation(pId, "width"), _dim[0]);
      glUniform1f(glGetUniformLocation(pId, "height"), _dim[1]);
      glUniform1i(glGetUniformLocation(pId, "inv"), 0);
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    }
  } else if(_currEffect == 4) {
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D,  tex[(nbIt4Effect4++) % 2],  0);
    glBindTexture(GL_TEXTURE_2D, tex[nbIt4Effect4 % 2]);
    glUniform1i(glGetUniformLocation(pId, "myTexture"), 0);
    glUniform1f(glGetUniformLocation(pId, "width"), _dim[0]);
    glUniform1f(glGetUniformLocation(pId, "height"), _dim[1]);
    glUniform1i(glGetUniformLocation(pId, "inv"), 0);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
  }
  
  glDisableVertexAttribArray(1);
  glDisableVertexAttribArray(0);
  glUseProgram(0);
  glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
  glBlitFramebuffer(0, 0, _dim[0], _dim[1], 0, 0, _dim[0], _dim[1], GL_COLOR_BUFFER_BIT, GL_NEAREST);
  glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

/*!\brief Cette fonction imprime le FPS (Frames Per Second) de
 * l'application.
 */
static void printFPS(void) {
  Uint32 t;
  static Uint32 t0 = 0, f = 0;
  f++;
  t = SDL_GetTicks();
  if(t - t0 > 2000) {
    fprintf(stderr, "%8.2f\n", (1000.0 * f / (t - t0)));
    t0 = t;
    f  = 0;
  }
}

static void myTex(GLuint * id) {
  glGenTextures(1, id);
  glBindTexture(GL_TEXTURE_2D, *id);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
}
